import numpy as np
cimport numpy as np
from libc.math cimport log10

from cython.parallel import prange, parallel
cimport openmp


ctypedef fused real_t:
    float
    double
    np.float32_t
    np.float64_t


def get_hist_from_weights_and_idigit(int num_bins, real_t[:] weights,
                                     long[:] i_digit):

    cdef int Np = weights.shape[0]

    cdef real_t[:] hist = np.zeros(num_bins+1, dtype=np.float64)

    cdef int ip, ib
    for ip in range(Np):
        ib = i_digit[ip]
        hist[ib] = hist[ib] + weights[ip]

    # Return a numpy array instead of a view
    tmp = np.zeros(num_bins-1, dtype=np.float64)
    tmp[:] = hist[1:num_bins]
    return tmp


def get_hist2d_from_weights(real_t [:] xvec, real_t [:] yvec,
                            real_t [:] weights,
                            real_t lower_x, real_t upper_x, int nbins_x,
                            real_t lower_y, real_t upper_y, int nbins_y,
                            bint logspace,
                            int numthreads=1):

    assert numthreads == 1, 'use project_image_omp for more than one thread'

    # Number of particles
    cdef int Np = xvec.shape[0]

    # Create hist2d array
    cdef real_t[:, :] hist2d = np.zeros((nbins_x, nbins_y),
                                        dtype=np.float64)

    # Loop integers and other variables
    cdef int ip, ix=0, iy=0
    cdef real_t x, y, dx, dy
    cdef real_t log10lower_x = log10(lower_x)
    cdef real_t log10lower_y = log10(lower_y)

    if logspace:
        dx = nbins_x/(log10(upper_x)-log10(lower_x))
        dy = nbins_y/(log10(upper_y)-log10(lower_y))
    else:
        dx = nbins_x/(upper_x-lower_x)
        dy = nbins_y/(upper_y-lower_y)

    for ip in range(Np):
        x = xvec[ip]
        y = yvec[ip]
        if (x > lower_x) and (x < upper_x) and (y > lower_y) and (y < upper_y):
            if logspace:
                ix = <int> ((log10(x) - log10lower_x)*dx)
                iy = <int> ((log10(y) - log10lower_y)*dy)
            else:
                ix = <int> ((x - lower_x)*dx)
                iy = <int> ((y - lower_y)*dy)

        if (ix >= 0) and (ix < nbins_x) and (iy >= 0) and (iy < nbins_y):
            hist2d[ix, iy] += weights[ip]

    # Fix to avoid returning a memory-view
    tmp = np.zeros((nbins_x, nbins_y), dtype=np.float64)
    tmp[:, :] = hist2d[:, :]

    return tmp


def get_hist2d_from_weights_omp(real_t [:] xvec, real_t [:] yvec,
                                real_t [:] weights,
                                real_t lower_x, real_t upper_x, int nbins_x,
                                real_t lower_y, real_t upper_y, int nbins_y,
                                bint logspace,
                                int numthreads=1):

    # Number of particles
    cdef int Np = xvec.shape[0]

    cdef int threadnum

    cdef int nx = nbins_x
    cdef int ny = nbins_y

    # Create hist2d array
    cdef real_t[:, :] hist2d = np.zeros((nbins_x, nbins_y),
                                        dtype=np.float64)
    cdef real_t[:, :, :] tmp_variable = np.zeros((nbins_x, nbins_y, numthreads),
                                                 dtype=np.float64)

    # Loop integers and other variables
    cdef int ip, ix=0, iy=0
    cdef real_t x, y, dx, dy
    cdef real_t log10lower_x = log10(lower_x)
    cdef real_t log10lower_y = log10(lower_y)

    if logspace:
        dx = nbins_x/(log10(upper_x)-log10(lower_x))
        dy = nbins_y/(log10(upper_y)-log10(lower_y))
    else:
        dx = nbins_x/(upper_x-lower_x)
        dy = nbins_y/(upper_y-lower_y)

    with nogil, parallel(num_threads=numthreads):
        for ip in prange(Np, schedule='static'):
            threadnum = openmp.omp_get_thread_num()

            x = xvec[ip]
            y = yvec[ip]
            if (x > lower_x) and (x < upper_x) and (y > lower_y) and (y < upper_y):
                if logspace:
                    ix = <int> ((log10(x) - log10lower_x)*dx)
                    iy = <int> ((log10(y) - log10lower_y)*dy)
                else:
                    ix = <int> ((x - lower_x)*dx)
                    iy = <int> ((y - lower_y)*dy)

            if (ix >= 0) and (ix < nbins_x) and (iy >= 0) and (iy < nbins_y):
                tmp_variable[ix, iy, threadnum] = tmp_variable[ix, iy, threadnum] + weights[ip]

    # Add up contributions from each thread
    for threadnum in range(numthreads):
        for ix in range(nbins_x):
            for iy in range(nbins_y):
                hist2d[ix, iy] = hist2d[ix, iy] + tmp_variable[ix, iy, threadnum]

    # Fix to avoid returning a memory-view
    tmp = np.zeros((nbins_x, nbins_y), dtype=np.float64)
    tmp[:, :] = hist2d[:, :]

    return tmp


def find_normalizing_norm_of_2d_hist(real_t [:, :] hist2d, real_t[:] edges_x,
                                     real_t[:] edges_y):

    cdef int ix, iy
    cdef real_t dx, dy, cell_area
    cdef real_t norm = 0

    cdef int nx = hist2d.shape[0]
    cdef int ny = hist2d.shape[1]
    for ix in range(nx):
        for iy in range(ny):
            dx = edges_x[ix+1] - edges_x[ix]
            dy = edges_y[iy+1] - edges_y[iy]
            cell_area = dx*dy
            norm += hist2d[ix, iy]*cell_area
    return norm



"""
Functions for getting derived variables of gas
"""
import numpy as np

# pylint: disable=import-outside-toplevel


def GFM_MetallicityTimesMasses(snap, get_dependencies=False):
    """Returns the metallicity times the mass"""
    if get_dependencies:
        return ['0_Masses', '0_GFM_Metallicity']
    return snap['0_Masses'] * snap['0_GFM_Metallicity']


def Volume(snap, get_dependencies=False):
    """Returns the volume of the Voronoi cells"""
    if get_dependencies:
        return ['0_Masses', '0_Density']
    return snap["0_Masses"] / snap["0_Density"]


def MachnumberTimesEnergyDissipation(snap, get_dependencies=False):
    """Returns the Mach number times the energy dissipation"""
    if get_dependencies:
        return ['0_Machnumber', '0_EnergyDissipation']
    variable = snap['0_Machnumber'] * snap['0_EnergyDissipation']
    return variable


def MagneticFieldSquared(snap, get_dependencies=False):
    """Returns the magnetic field strength squared"""
    if get_dependencies:
        return ['0_MagneticField']
    return np.sum(snap['0_MagneticField']**2, axis=1)


def MagneticFieldStrength(snap, get_dependencies=False):
    """Returns the magnetic field strength"""
    if get_dependencies:
        return ['0_MagneticField']
    return np.sqrt(np.sum(snap['0_MagneticField']**2, axis=1))


def VelocityMagnitude(snap, get_dependencies=False):
    """Returns the magnitude of the gas velocity"""
    if get_dependencies:
        return ['0_Velocities']
    return np.sqrt(np.sum(snap['0_Velocities']**2, axis=1))


def MagneticFieldSquaredTimesVolume(snap, get_dependencies=False):
    """Returns B² × V """
    if get_dependencies:
        return ['0_Volume', '0_MagneticField']
    variable = snap["0_Volume"] * np.sum(snap['0_MagneticField']**2, axis=1)
    return variable


def Pressure(snap, get_dependencies=False):
    """Returns the gas pressure"""
    if get_dependencies:
        return ['0_InternalEnergy', '0_Density']

    if snap.gamma == 1:
        msg = 'Temperature field not supported for isothermal EOS!'
        raise RuntimeError(msg)
    gm1 = snap.gamma - 1
    variable = snap["0_InternalEnergy"] * snap["0_Density"] * gm1
    return variable


def PressureTimesVolume(snap, get_dependencies=False):
    """Returns the gas pressure times the volume"""
    if get_dependencies:
        return ['0_Pressure', '0_Volume']

    if '0_Pressure' in snap:
        return snap['0_Pressure'] * snap['0_Volume']

    if snap.gamma != 1:
        gm1 = snap.gamma - 1
        variable = snap["0_Masses"] * snap["0_InternalEnergy"] * gm1
    else:
        variable = snap['0_Volume'] * snap['0_Pressure']

    return variable


def Temperatures(snap, get_dependencies=False):
    """Returns the temperature in Kelvin"""

    if get_dependencies:
        return ['0_InternalEnergy', '0_MeanMolecularWeight']

    from astropy import constants as c
    mhydrogen = c.m_e + c.m_p

    gm1 = snap.gamma - 1

    if snap.gamma == 1:
        msg = 'Temperature field not supported for isothermal EOS!'
        raise RuntimeError(msg)

    mmean = snap['0_MeanMolecularWeight']

    # temperature in Kelvin
    from .. import settings
    if settings.use_units:
        variable = (gm1 * snap["0_InternalEnergy"]
                    * mmean * mhydrogen).to('K')
    else:
        u_v = snap.arepo_units['unit_velocity']
        variable = (gm1 * snap["0_InternalEnergy"]
                    * u_v**2 * mmean * mhydrogen
                    ).to('K').value
    return variable


def TemperaturesTimesMasses(snap, get_dependencies=False):
    """Returns the temperature times masses"""
    if get_dependencies:
        return ['0_Temperatures', '0_Masses']

    return snap["0_Temperatures"] * snap['0_Masses']


def Current(snap, get_dependencies=False):
    """Returns the magnitude of the current, i.e., |∇×B|"""

    if get_dependencies:
        return ['0_BfieldGradient']

    def get_index(ii, jj):
        return ii * 3 + jj
    gradB = snap['0_BfieldGradient']
    J_x = gradB[:, get_index(2, 1)] - gradB[:, get_index(1, 2)]
    J_y = gradB[:, get_index(0, 2)] - gradB[:, get_index(2, 0)]
    J_z = gradB[:, get_index(1, 0)] - gradB[:, get_index(0, 1)]

    J = np.sqrt(J_x**2 + J_y**2 + J_z**2)
    return J


def Enstrophy(snap, get_dependencies=False):
    """Returns the enstrophy, i.e., 1/2|∇×v|²"""

    if get_dependencies:
        return ['0_VelocityGradient']

    def get_index(ii, jj):
        return ii * 3 + jj
    gradV = snap['0_VelocityGradient'][()]
    vor_x = gradV[:, get_index(2, 1)] - gradV[:, get_index(1, 2)]
    vor_y = gradV[:, get_index(0, 2)] - gradV[:, get_index(2, 0)]
    vor_z = gradV[:, get_index(1, 0)] - gradV[:, get_index(0, 1)]

    enstrophy = 0.5 * (vor_x**2 + vor_y**2 + vor_z**2)
    return enstrophy


def EnstrophyTimesMasses(snap, get_dependencies=False):
    """ Returns the enstrophy times masses"""

    if get_dependencies:
        return ['0_VelocityGradient']

    # Reshaping is slow
    if False:
        n_cells = snap['0_VelocityGradient'].shape[0]

        # Reshape to tensor form
        gradV = snap['0_VelocityGradient'].reshape(n_cells, 3, 3)
        # Get vorticity components
        vor_x = gradV[:, 2, 1] - gradV[:, 1, 2]
        vor_y = gradV[:, 0, 2] - gradV[:, 2, 0]
        vor_z = gradV[:, 1, 0] - gradV[:, 0, 1]
    else:
        def get_index(ii, jj):
            return ii * 3 + jj
        gradV = snap['0_VelocityGradient']
        vor_x = gradV[:, get_index(2, 1)] - gradV[:, get_index(1, 2)]
        vor_y = gradV[:, get_index(0, 2)] - gradV[:, get_index(2, 0)]
        vor_z = gradV[:, get_index(1, 0)] - gradV[:, get_index(0, 1)]
    # The vorticity vector
    # vorticity = np.stack([vor_x, vor_y, vor_z], axis=1)

    enstrophy = 0.5 * (vor_x**2 + vor_y**2 + vor_z**2)
    variable = enstrophy * snap['0_Masses']

    return variable


def MeanMolecularWeight(snap, get_dependencies=False):
    """Returns the mean molecular weight, μ"""

    if get_dependencies:
        return ['0_Density']

    if '0_GFM_Metals' in snap.info(0, False):
        hydrogen_abundance = snap['0_GFM_Metals'][:, 0]
    else:
        hydrogen_abundance = np.array([0.76])

    if '0_ElectronAbundance' in snap.info(0, False):
        electron_abundance = snap['0_ElectronAbundance']
        # partially ionized
        mean_molecular_weight = 4. / (1. + 3. * hydrogen_abundance
                                      + 4. * hydrogen_abundance
                                      * electron_abundance)
    else:
        # fully ionized
        mean_molecular_weight = 4. / (5. * hydrogen_abundance + 3.)
    return mean_molecular_weight


def NumberDensity(snap, get_dependencies=False):
    """
    The gas number density in cm⁻³.
    """
    if get_dependencies:
        return ['0_Density']

    from astropy import constants as c
    density = snap['0_Density'].cgs
    mean_molecular_weight = snap['0_MeanMolecularWeight']
    proton_mass = c.m_p.to('g')
    number_density_gas = density / (mean_molecular_weight * proton_mass)
    return number_density_gas


def MagneticCurvature(snap, get_dependencies=False):
    """Returns the length of the magnetic curvature vector"""

    if get_dependencies:
        return ['0_MagneticField', '0_BfieldGradient']

    from .. import util

    @util.remove_astro_units
    def get_func(B, gradB):
        from ..cython.get_derived_variables import get_curvature
        return get_curvature(B, gradB)

    curva = get_func(snap['0_MagneticField'], snap['0_BfieldGradient'])
    unit_quantity = snap['0_BfieldGradient'].uq / snap['0_MagneticField'].uq
    curva = curva * unit_quantity
    return curva


def VelocityCurvature(snap, get_dependencies=False):
    """Returns the length of the velocity curvature vector"""
    if get_dependencies:
        return ['0_Velocities', '0_VelocityGradient']

    from .. import util

    @util.remove_astro_units
    def get_func(V, gradV):
        from ..cython.get_derived_variables import get_curvature
        return get_curvature(V, gradV)

    curva = get_func(snap['0_Velocities'], snap['0_VelocityGradient'])
    unit_quantity = snap['0_VelocityGradient'].uq / snap['0_Velocities'].uq
    curva = curva * unit_quantity
    return curva


def Diameters(snap, get_dependencies=False):
    """Returns the cell diameter"""
    if get_dependencies:
        return ['0_Volume']

    return 2 * np.cbrt((snap['0_Volume']) / (4.0 * np.pi / 3.0))


functions = {
    "0_GFM_MetallicityTimesMasses": GFM_MetallicityTimesMasses,
    "0_Volume": Volume,
    "0_Temperatures": Temperatures,
    "0_MachnumberTimesEnergyDissipation": MachnumberTimesEnergyDissipation,
    "0_MagneticFieldSquared": MagneticFieldSquared,
    "0_MagneticFieldStrength": MagneticFieldStrength,
    "0_MagneticFieldSquaredTimesVolume": MagneticFieldSquaredTimesVolume,
    "0_Pressure": Pressure,
    "0_PressureTimesVolume": PressureTimesVolume,
    "0_TemperaturesTimesMasses": TemperaturesTimesMasses,
    "0_Current": Current,
    "0_Enstrophy": Enstrophy,
    "0_EnstrophyTimesMasses": EnstrophyTimesMasses,
    "0_MeanMolecularWeight": MeanMolecularWeight,
    "0_NumberDensity": NumberDensity,
    "0_MagneticCurvature": MagneticCurvature,
    "0_VelocityMagnitude": VelocityMagnitude,
    "0_VelocityCurvature": VelocityCurvature,
    "0_Diameters": Diameters
}



"""
Defines a class that creates an image of a given variable by projecting it
onto a 2D plane.
"""
import numpy as np
from .image_creator import ImageCreator
from .. import util
from .. import settings
from .. import units


class Projector(ImageCreator):
    """
    A class that allows creating an image of a given variable by projecting
    it onto a 2D plane.

    The Projector class is a subclass of the ImageCreator class.
    The Projector class creates an image of a given variable by projecting
    it onto a 2D plane.

    It takes in several parameters such as a snapshot object, center and
    widths of the region, direction of projection, and various optional
    parameters for number of pixels, smoothing length.
    """

    def __init__(self, snap, center, widths, direction,
                 npix=512, parttype=0, nvol=8, make_snap_with_selection=True):
        """
        Initialize the Projector class.

        Parameters
        ----------
        snap : Snapshot
            A snapshot object of Snapshot class from paicos package.

        center : numpy array
            Center of the region on which projection is to be done, e.g.
            center = [x_c, y_c, z_c].

        widths : numpy array
            Widths of the region on which projection is to be done,
            e.g.m widths=[width_x, width_y, width_z].

        direction : str
            Direction of the projection, e.g. 'x', 'y' or 'z'.

        npix : int, optional
            Number of pixels in the horizontal direction of the image,
            by default 512.

        parttype : int, optional
            Number of the particle type to project, by default gas (PartType 0).

        nvol : int, optional
            Integer used to determine the smoothing length, by default 8

        """

        # call the superclass constructor to initialize the ImageCreator class
        super().__init__(snap, center, widths, direction, npix=npix, parttype=parttype)

        # nvol is an integer that determines the smoothing length
        self.nvol = nvol

        self._observers = []

        self.make_snap_with_selection = make_snap_with_selection

        # Call selection
        self.has_do_region_selection_been_called = False
        self._do_region_selection()
        self.has_do_region_selection_been_called = True

    def _bind_to(self, callback):
        self._observers.append(callback)

    def _do_region_selection(self):

        self.do_unit_consistency_check()

        if self.has_do_region_selection_been_called:
            if self.make_snap_with_selection:
                err_msg = ("It looks like you are changing projector "
                           + "properties after the fact, i.e. changing widths "
                           + "center, orientation, resolution etc. This does not "
                           + "work with the option make_snap_with_selection, which "
                           + "you have turned on.")
                raise RuntimeError(err_msg)
        center = self.center
        widths = self.widths
        snap = self.snap
        parttype = self.parttype

        # get the index of the region of projection
        if self.direction != 'orientation':
            get_index = util.get_index_of_cubic_region
            self.index = get_index(self.snap[f"{parttype}_Coordinates"],
                                   center, widths, snap.box)
        else:
            get_index = util.get_index_of_rotated_cubic_region
            self.index = get_index(snap[f"{parttype}_Coordinates"],
                                   center, widths, snap.box,
                                   self.orientation)

        # Reduce the snapshot to only contain region of interest
        if self.make_snap_with_selection:
            self.snap = self.snap.select(self.index, parttype=parttype)

        # Calculate the smoothing length
        avail_list = (list(snap.keys()) + snap._auto_list)
        if f'{parttype}_Volume' in avail_list:
            self.hsml = np.cbrt(self.nvol * (self.snap[f"{parttype}_Volume"])
                                / (4.0 * np.pi / 3.0))
        elif f'{parttype}_SubfindHsml' in avail_list:
            self.hsml = self.snap[f'{parttype}_SubfindHsml']
        else:
            raise RuntimeError(
                'There is no smoothing length or volume for the projector')

        self.pos = self.snap[f'{self.parttype}_Coordinates']

        if settings.use_units:
            self.hsml = self.hsml.to(self.pos.unit)

        if not self.make_snap_with_selection:
            self.hsml = self.hsml[self.index]
            self.pos = self.pos[self.index]

        # Call other functions that need to be updated
        for callback in self._observers:
            # print(callback, 'from projector')
            callback()

    @util.remove_astro_units
    def _cython_project(self, center, widths, variable):
        """
        Private method for projecting using cython
        """
        if settings.openMP_has_issues:
            from ..cython.sph_projectors import project_image as project
            from ..cython.sph_projectors import project_oriented_image as project_orie
        else:
            from ..cython.sph_projectors import project_image_omp as project
            from ..cython.sph_projectors import project_oriented_image_omp as project_orie

        x_c, y_c, z_c = center[0], center[1], center[2]
        width_x, width_y, width_z = widths

        boxsize = self.snap.box
        if self.direction == 'x':
            projection = project(self.pos[:, 1],
                                 self.pos[:, 2],
                                 variable,
                                 self.hsml, self.npix,
                                 y_c, z_c, width_y, width_z,
                                 boxsize, settings.numthreads_reduction)
        elif self.direction == 'y':
            projection = project(self.pos[:, 2],
                                 self.pos[:, 0],
                                 variable,
                                 self.hsml, self.npix,
                                 z_c, x_c, width_z, width_x,
                                 boxsize, settings.numthreads_reduction)
        elif self.direction == 'z':
            projection = project(self.pos[:, 0],
                                 self.pos[:, 1],
                                 variable,
                                 self.hsml, self.npix,
                                 x_c, y_c, width_x, width_y,
                                 boxsize, settings.numthreads_reduction)
        elif self.direction == 'orientation':
            unit_vectors = self.orientation.cartesian_unit_vectors

            projection = project_orie(self.pos[:, 0],
                                      self.pos[:, 1],
                                      self.pos[:, 2],
                                      variable,
                                      self.hsml, self.npix,
                                      x_c, y_c, z_c, width_x, width_y,
                                      boxsize,
                                      unit_vectors['x'],
                                      unit_vectors['y'],
                                      unit_vectors['z'],
                                      settings.numthreads_reduction)

        return projection

    def project_variable(self, variable):
        """
        projects a given variable onto a 2D plane.

        Parameters
        ----------
        variable : str, function, numpy array
            variable, it can be passed as string or an array

        Returns
        -------
        numpy array
            The image of the projected variable
        """
        self.do_unit_consistency_check()
        # This calls _do_region_selection if resolution, Orientation,
        # widths or center changed
        self._check_if_properties_changed()

        if isinstance(variable, str):
            err_msg = 'projector uses a different parttype'
            assert int(variable[0]) == self.parttype, err_msg
            variable = self.snap[variable]
        else:
            if not isinstance(variable, np.ndarray):
                raise RuntimeError('Unexpected type for variable')

        if variable.shape == self.index.shape:
            variable = variable[self.index]

        assert len(variable.shape) == 1, 'only scalars can be projected'

        # Do the projection
        projection = self._cython_project(self.center, self.widths, variable)

        # Transpose
        projection = projection.T

        assert projection.shape[1] == self.npix_width, (projection.shape,
                                                        self.npix_width)

        assert projection.shape[0] == self.npix_height, (projection.shape,
                                                         self.npix_height)

        area_per_pixel = self.area / np.prod(projection.shape)

        if isinstance(variable, units.PaicosQuantity):
            projection = projection * variable.unit_quantity

        return projection / area_per_pixel



"""
Defines a class that creates an image of a given variable by finding
the Voronoi cells closest to the image plane.
"""
import numpy as np
from scipy.spatial import KDTree
from .image_creator import ImageCreator
from .. import util
from .. import settings
from ..orientation import Orientation


class Slicer(ImageCreator):
    """
    This implements slicing of gas variables.
    """

    def __init__(self, snap, center, widths, direction,
                 npix=512, parttype=0, make_snap_with_selection=False):
        """
        Initialize the Slicer object.

        Parameters
        ----------
        snap : Snapshot
            A snapshot object of Snapshot class from paicos package.

        center :
            Center of the region on which slicing is to be done, e.g.
            center = [x_c, y_c, z_c].

        widths :
            Widths of the region on which slicing is to be done,
            e.g. widths=[width_x, width_y, width_z] where one of the widths
            is zero (e.g. width_x=0 if direction='x').

        direction : str
            Direction of the slicing, e.g. 'x', 'y' or 'z'. For instance,
            setting direction to 'x' gives a slice in the yz plane with the
            constant x-value set to be x_c.

        npix : int, optional
            Number of pixels in the horizontal direction of the image,
            by default 512.

        parttype : int, optional
            The particle type to project, by default gas (PartType 0).

        make_snap_with_selection : bool
            a boolean indicating if a new snapshot object should be made with
            the selected region, defaults to False

        """

        if make_snap_with_selection:
            raise RuntimeError('make_snap_with_selection not yet implemented!')

        super().__init__(snap, center, widths, direction, npix=npix, parttype=parttype)

        for ii, direc in enumerate(['x', 'y', 'z']):
            if self.direction == direc:
                assert self.widths[ii] == 0.
        if self.direction == 'orientation':
            assert self.widths[2] == 0.

        self._do_region_selection()

    def _do_region_selection(self):

        self.do_unit_consistency_check()

        parttype = self.parttype
        snap = self.snap
        center = self.center
        widths = self.widths

        # Pre-select a narrow region around the region-of-interest
        avail_list = (list(snap.keys()) + snap._auto_list)
        if f'{parttype}_Volume' in avail_list:
            thickness = 4.0 * np.cbrt((snap[f"{parttype}_Volume"])
                                      / (4.0 * np.pi / 3.0))
        elif f'{parttype}_SubfindHsml' in avail_list:
            thickness = snap[f'{parttype}_SubfindHsml']
        else:
            raise RuntimeError(
                'There is no smoothing length or volume for the thickness of the slice')

        if hasattr(thickness, 'unit'):
            thickness = thickness.to(center.unit)

        if self.direction != 'orientation':
            get_index = util.get_index_of_cubic_region_plus_thin_layer
            self.slice = get_index(snap[f"{parttype}_Coordinates"],
                                   center, widths, thickness,
                                   snap.box)
        else:
            get_index = util.get_index_of_rotated_cubic_region_plus_thin_layer
            self.slice = get_index(snap[f"{parttype}_Coordinates"],
                                   center, widths, thickness, snap.box,
                                   self.orientation)

        self.index_in_slice_region = np.arange(snap[f"{parttype}_Coordinates"].shape[0]
                                               )[self.slice]

        # Construct a tree
        self.pos = snap[f"{parttype}_Coordinates"][self.slice]
        tree = KDTree(self.pos)

        # Now construct the image grid
        w, h = self._get_width_and_height_arrays()

        center = self.center
        ones = np.ones(w.shape[0])
        if self.direction == 'x':
            image_points = np.vstack([ones * center[0], w, h]).T
        elif self.direction == 'y':
            image_points = np.vstack([h, ones * center[1], w]).T
        elif self.direction == 'z':
            image_points = np.vstack([w, h, ones * center[2]]).T
        elif self.direction == 'orientation':
            orientation = self.orientation
            image_points = np.vstack([w, h, ones * center[2]]).T - self.center
            image_points = np.matmul(orientation.rotation_matrix, image_points.T).T \
                + self.center
        else:
            raise RuntimeError(f"Problem with direction={self.direction} input")

        # Query the tree to obtain closest Voronoi cell indices
        d, i = tree.query(image_points, workers=settings.numthreads)

        self.index = self._unflatten(self.index_in_slice_region[i])
        self.distance_to_nearest_cell = self._unflatten(d)

    def _get_width_and_height_arrays(self):
        """
        Get width and height coordinates in the image as 1D arrays
        of total length npix_width × npix_height.
        """

        # TODO: Make this part of the image_creator class

        extent = self.extent
        npix_width = self.npix_width
        npix_height = self.npix_height
        width = self.width
        height = self.height

        w = extent[0] + (np.arange(npix_width) + 0.5) * width / npix_width
        h = extent[2] + (np.arange(npix_height) + 0.5) * height / npix_height

        if settings.use_units:
            wu = w.unit_quantity
            ww, hh = np.meshgrid(w.value, h.value)
            ww = ww * wu
            hh = hh * wu
        else:
            ww, hh = np.meshgrid(w, h)

        w = ww.flatten()
        h = hh.flatten()

        np.testing.assert_array_equal(ww, self._unflatten(ww.flatten()))

        return w, h

    def _unflatten(self, arr):
        """
        Helper function to un-flatten 1D arrays to a 2D image
        """
        return arr.flatten().reshape((self.npix_height, self.npix_width))

    def slice_variable(self, variable):
        """
        Slice a gas variable based on the Voronoi cells closest to the image
        plane.

        Parameters
        ----------
        variable: a string or an array of shape (N, )
                  representing the gas variable to slice

        Returns:
        An array of shape (npix, npix) representing the sliced gas variable
        """

        # This calls _do_region_selection if resolution, Orientation,
        # widths or center changed
        self._check_if_properties_changed()

        if isinstance(variable, str):
            err_msg = 'slicer uses a different parttype'
            assert int(variable[0]) == self.parttype, err_msg
            variable = self.snap[variable]
        else:
            if not isinstance(variable, np.ndarray):
                raise RuntimeError('Unexpected type for variable')

        return variable[self.index]

    @property
    def depth(self):
        if self.direction == 'x':
            return self.width_x

        elif self.direction == 'y':
            return self.width_y

        elif self.direction == 'z' or self.direction == 'orientation':
            return self.width_z

    @depth.setter
    def depth(self, value):
        err_msg = "You can't change the depth of a slicer, as it is zero by definition"
        raise RuntimeError(err_msg)
